# coding=utf-8
from __future__ import annotations

import enum
import types
from typing import TYPE_CHECKING, Optional, Type  # noqa: F401

import depthai as dai
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, create_model, BeforeValidator
from typing_extensions import TypedDict, Annotated

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from pydantic.fields import FieldInfo


def create_typed_dict(name: str, keys: dict[str, type]):
    """
    Create a TypedDict class with the given name and keys.

    用于创建一个类型字典

    Args:
        name: 类型字典名称
        keys: 包含类型字典键值对的字典

    Returns:
        Type[TypedDict]: 返回一个类型字典
    """
    # 使用 types 模块的 new_class 函数创建一个类，
    # 类名为 name，父类为 TypedDict，
    # kwds 为 total=False，
    # exec_body 为一个函数，用于更新__annotations__属性
    return types.new_class(
        name,
        (TypedDict,),
        kwds={"total": False},
        exec_body=lambda ns: ns.update(__annotations__=keys),
    )


def create_pydantic_model(name: str, keys: dict[str, tuple[type, FieldInfo]]):
    """
    Create a Pydantic model with the given name and keys.

    用于创建一个 Pydantic 模型

    Args:
        name: 模型名称
        keys: 包含模型字段名和类型的字典

    Returns:
        Type[BaseModel]: 返回一个 Pydantic 模型
    """
    return create_model(name, **keys)


def create_enum(name: str, keys: str | Iterator | Mapping, enum_type: type | None = None):
    """
    Create an enum with the given name and keys.

    用于创建一个枚举

    Args:
        name: 枚举名称
        keys: 枚举值
        enum_type: 枚举类型

    Returns:
        Type[Enum]: 返回一个枚举
    """
    # 使用 enum 模块的 Enum 函数创建一个枚举，枚举名称为 name，枚举值为 keys，类型为 enum_type
    return enum.Enum(
        name,
        keys,
        type=enum_type,
    )


# 获取相机接口的枚举
camera_board_socket = list(dai.CameraBoardSocket.__members__)[1:]

# 创建一个枚举，用于获取相机传感器类型
CameraSensorType = create_enum(
    "CameraSensorType",
    {SensorType: SensorType.lower() for SensorType in dai.CameraSensorType.__members__},
    str,
)

# 创建一个枚举，用于获取相机接口
CameraBoardSocket = create_enum(
    "CameraBoardSocket",
    {socket: socket for socket in camera_board_socket},
    str,
)

# 定义默认配置
default_config = ConfigDict(
    # 是否去除字符串两端的空白字符
    str_strip_whitespace=True,
    # 是否使用枚举值
    use_enum_values=True,
    # 是否验证赋值
    validate_assignment=True,
    # 是否验证默认值
    validate_default=True,
    # 是否将数字转换为字符串
    coerce_numbers_to_str=True,
)


# 定义一个模型，用于相机之间的平移值
class SpecTranslation(BaseModel):
    model_config = default_config

    x: float = Field(
        default=0.0,
        description="Translation along the x axis (cm)",
    )
    y: float = Field(
        default=0.0,
        description="Translation along the y axis (cm)",
    )
    z: float = Field(
        default=0.0,
        description="Translation along the z axis (cm)",
    )


# 定义一个模型，用于相机之间的旋转值
class SpecRotation(BaseModel):
    model_config = default_config

    r: float = Field(
        default=0.0,
        description="Rotation around the x axis",
    )
    p: float = Field(
        default=0.0,
        description="Rotation around the y axis",
    )
    y: float = Field(
        default=0.0,
        description="Rotation around the z axis",
    )


class Extrinsics(BaseModel):
    model_config = default_config

    to_cam: CameraBoardSocket | None = Field(
        default=None,
        description="Name of the camera to which the extrinsic apply",
        validation_alias=AliasChoices("to_cam", "toCameraSocket"),
        serialization_alias="to_cam",
    )
    spec_translation: SpecTranslation | None = Field(
        default=None,
        description="Translation of the camera in the spec frame",
        validation_alias=AliasChoices("specTranslation", "spec_translation"),
        serialization_alias="specTranslation",
    )
    rotation: SpecRotation | None = Field(
        default=None,
        description="Rotation of the camera in the spec frame",
        serialization_alias="rotation",
    )


strFloat = Annotated[float, BeforeValidator(float)]


class Camera(BaseModel):
    model_config = default_config

    name: str | None = Field(
        default=None,
        description="Name of the camera",
        serialization_alias="name",
    )
    h_fov: float = Field(
        default=0.0,
        description="Horizontal field of view",
        ge=0.0,
        validation_alias=AliasChoices("hfov", "horizontalFov", "h_fov"),
        serialization_alias="hfov",
    )
    cam_type: CameraSensorType = Field(
        default="color",
        description="Type of camera",
        validation_alias=AliasChoices("cam_type", "cameraType", "type"),
        serialization_alias="type",
    )
    extrinsics: Extrinsics | None = Field(
        default=None,
        description="Extrinsic parameters of the camera",
        serialization_alias="extrinsics",
    )


# CameraDict = create_typed_dict(
#     "CameraDict",
#     {socket: Camera for socket in camera_board_socket},
# )

CameraDict = create_pydantic_model(
    "CameraDict",
    {socket: (Optional[Camera], Field(None)) for socket in camera_board_socket},
)


class StereoConfig(BaseModel):
    model_config = default_config

    left_cam: CameraBoardSocket | None = Field(
        default=None,
        description="Name of the left camera",
        serialization_alias="left_cam",
    )
    right_cam: CameraBoardSocket | None = Field(
        default=None,
        description="Name of the right camera",
        serialization_alias="right_cam",
    )


class BoardConfig(BaseModel):
    model_config = default_config

    name: str = Field(
        default_factory=str,
        description="Name of the board",
        validation_alias=AliasChoices("name", "productName"),
        serialization_alias="name",
    )
    revision: str = Field(
        default_factory=str,
        description="Revision of the board",
        validation_alias=AliasChoices("revision", "boardRev"),
        serialization_alias="revision",
    )
    cameras: CameraDict | None= Field(
        default=None,
        # default_factory=CameraDict,
        description="Camera configuration",
        serialization_alias="cameras",
    )
    stereo_config: StereoConfig | None = Field(
        default=None,
        description="Stereo camera configuration",
        serialization_alias="stereo_config",
    )


class CalibrationModel(BaseModel):
    model_config = default_config

    board_config: BoardConfig = Field(
        default_factory=BoardConfig,
        description="Board configuration",
        serialization_alias="board_config",
    )


CalibrationModel().model_copy(update={})
