# coding=utf-8
from __future__ import annotations

import depthai as dai
from loguru import logger
from rich import print

from calibration_model import (
    CalibrationModel,
    Camera,
    CameraDict,
    Extrinsics,
    SpecRotation,
    SpecTranslation,
    StereoConfig,
)

with dai.Device() as device:
    device_name = device.getDeviceName()
    try:
        calib: dai.CalibrationHandler = device.readCalibration2()
        print("Read User Calibration")
    except dai.EepromError:
        calib: dai.CalibrationHandler = device.readFactoryCalibration()
        print("Read Factory Calibration")

    eeprom_data: dai.EepromData = calib.getEepromData()

    a_calibration_model = CalibrationModel()

    board_config = a_calibration_model.board_config
    cameras = board_config.cameras or CameraDict()
    stereo_config = board_config.stereo_config or StereoConfig()

    prod_name = eeprom_data.productName
    board_name = eeprom_data.boardName
    board_rev = eeprom_data.boardRev

    left_cam = calib.getStereoLeftCameraId()
    right_cam = calib.getStereoRightCameraId()

    board_config.name = device_name or prod_name or board_name
    board_config.revision = board_rev

    stereo_config.left_cam = left_cam.name
    stereo_config.right_cam = right_cam.name

    cams: list[dai.CameraFeatures] = device.getConnectedCameraFeatures()

    for i, cam in enumerate(cams):
        socket = cam.socket
        socket_name = socket.name
        try:
            h_fov = calib.getFov(socket, useSpec=True)
        except:
            logger.exception("Failed to get FOV for camera %s", socket_name)
            h_fov = 0.0

        cam_type = cam.supportedTypes[0].name.lower()

        try:
            to_cam = cams[i + 1].socket.name
        except:
            to_cam = None

        extrinsics = (
            Extrinsics(to_cam=to_cam, spec_translation=SpecTranslation(), rotation=SpecRotation())
            if to_cam is not None
            else None
        )

        if socket_name == "CAM_A":
            extrinsics = None

        cameras = cameras.model_copy(
            update={
                socket_name: Camera(
                    name=cam.name,
                    h_fov=f"{h_fov:.2f}",
                    cam_type=cam_type,
                    extrinsics=extrinsics,
                )
            }
        )

    board_config.cameras = cameras
    board_config.stereo_config = stereo_config

    export_json = a_calibration_model.model_dump_json(
        indent=2,
        by_alias=True,
        exclude_none=True,
    )
    print(f"{export_json}")
    with open(f"{board_name}.json", "w", encoding="utf8") as f:
        f.write(export_json)
